//Start with a led
#define startPin 22 //pin for the start signal
int startState = 0; //variable for the state of the start signal
#define startLed 24 //pin for the led of the start

#define secCheckPin 26 //pin for the security window over the spindle
int secCheckState = 0; //variable for the state of the security window over the spindle
#define pauseLed 28 //pin for the led of the spindle window

bool stopState = 0;
#define stopLed 32

#define emergencyStopPin 23 //pin of the button to enable the interrupt incase of an emergency stop of the machine
bool emergencyState = 0;
#define emergencyLed 44

bool receptionState = 0;

bool momentaryState = 0;

//LCD display
#include <LiquidCrystal.h>
const int rs=50, en=52, d4=53, d5=51, d6=49, d7=47;
LiquidCrystal lcdSpeed(rs, en, d4, d5, d6, d7);

int targetSpeed = 0; //target speed of the DC motor [rpm]
float tolerance = 0.0;
int MinimumTargetSpeed = 0;
int MaximumTargetSpeed = 0;
int okSpeed = 0;
bool receiveOK = 0;
int checkSpeed[100];
int i = 0;
int j = 0;
int realSpeed = 0;

int enableState = 0;

// Potentiometer reading for the speed of the spindle
#define pwmPin 6 //pin for the PWM signal (frequency = 488 Hz and the default frequency is 31250 Hz)
int potValue = 0; //variable for the acquisition of the signal from the potentiometer
int newPotValue = 0; //variable for the map of the potentiometer signal

// rotary encoder signals for the motor speed
#include <util/atomic.h>
#define ENCA 2 //signal A
#define ENCB 3 //signal B

//global variables to store the number of counts from the encoder
volatile int b = 0;
volatile int increment = 0;
int pos = 0;
float deltaT = 0;
float vel = 0;
float velocity = 0;
long currT = 0;
long prevT = 0;
int posPrev = 0;
volatile int pos_i = 0;
float velFilt = 0; //filtered output speed at instant k
float velFilt1 = 0; //filtered output speed at instant k-1
float velPrev1 = 0; //unfiltered output speed at instant k-1
int velFiltInt = 0;

//PID controller
float kp = 0; //proportional constant
float ki = 0; //integrative constant
float kd = 0; //derivative constant

float error = 0;
float eintegral = 0;
float ederivative = 0;
float lastError = 0;

float LowerLimit_Iout = 0;
float UpperLimit_Iout = 255;

float Pout = 0;
float Iout = 0;
float Dout = 0;
float u = 0;
int pwr = 0;

volatile int spd = 0;
bool snd = 0;

int enb = 0;
int enb_rot = 0;
int enb_spd = 0;
int enb_stop = 0;
int halfSpeed = 0;

float sampleTime = 0.1;

void setup(){
  Serial.begin(9600);
  Serial3.begin(9600);

  //enabling of the pins for the reading of the start and PWM signals
  pinMode(startPin,INPUT_PULLUP);
  pinMode(startLed,OUTPUT);

  pinMode(secCheckPin,INPUT_PULLUP);
  pinMode(pauseLed,OUTPUT);

  pinMode(stopLed,OUTPUT);

  pinMode(emergencyStopPin,INPUT_PULLUP);
  pinMode(emergencyLed,OUTPUT);
  
  pinMode(pwmPin,OUTPUT);
  setPwmFrequency(pwmPin,1); //set the pin and the divisor on the timer for the PWM signal

  pinMode(ENCA,INPUT_PULLUP);
  pinMode(ENCB,INPUT_PULLUP);

  //interrupt connected to the pin n°2 to trigger when the signal rises
  attachInterrupt(digitalPinToInterrupt(ENCA),readEncoder,RISING);
  
  lcdSpeed.begin(16,2);
  lcdSpeed.clear();

  delay(3000);
}

void loop(){
  lcdSpeed.clear();
  digitalWrite(startLed,LOW); //keep the start led off (ON and window close), green
  digitalWrite(pauseLed,LOW); //keep the pause led off (ON and window open), blue
  digitalWrite(stopLed,LOW); //turn on the yellow-stop light
  digitalWrite(emergencyLed,LOW);
  
  startState = digitalRead(startPin); //acquisition of the signal from the start
  secCheckState = digitalRead(secCheckPin); //acquisition of the signal from the securuty window over the spindle
  
  if(startState == HIGH && secCheckState == LOW){ //switch on the machine and the window must be open
    setMotor(0,pwmPin);

    enb = receiveEnableStart();

    if(startState == HIGH && secCheckState == LOW && enb == 1){
      digitalWrite(startLed,LOW); //keep the start led off (ON and window close), green
      digitalWrite(pauseLed,LOW); //keep the pause led off (ON and window open), blue
      digitalWrite(stopLed,HIGH); //turn on the orange-stop light

      lcdSpeed.clear();
      lcdSpeed.print("Machine ready");
      lcdSpeed.setCursor(0,1);
      lcdSpeed.print("for a process");

      snd = 0; //variable useful for the transmission of the speed

      targetSpeed = 0;
      do{
        startState = digitalRead(startPin); //acquisition of the signal from the start
        targetSpeed = receiveSpeed();
        if(receiveOK == 2){        
          return;
        }else if(startState == LOW){
          Emergency();
          return;
        }
      }while(targetSpeed == 0 && startState == HIGH && receiveOK == 0);
      tolerance = 0.02;
      MinimumTargetSpeed = targetSpeed - tolerance*targetSpeed;
      MaximumTargetSpeed = targetSpeed + tolerance*targetSpeed;
      Serial.println(targetSpeed);

      digitalWrite(startLed,LOW); //switch on the start led (ON and window closed)
      digitalWrite(pauseLed,HIGH); //when the blue light switches on, the window can be closed
      digitalWrite(stopLed,HIGH);

      lcdSpeed.clear();
      lcdSpeed.print("Close the window");
      lcdSpeed.setCursor(0,1);
      lcdSpeed.print("to start process");

      do{
        secCheckState = digitalRead(secCheckPin);
        startState = digitalRead(startPin);
      }while(secCheckState == LOW && startState == HIGH);
      if(startState == LOW){
        Emergency();
        return;
      }

      enableState = 0;
      do{
        enableState = receiveEnableRotation();
      }while(enableState == 0);
      if(startState == LOW || secCheckState == LOW){
        Emergency();
        return;
      }

      if(startState == HIGH && secCheckState == HIGH && enableState == 1){
        i = 0;
        j = 0;
        realSpeed = 0;
        velFiltInt = 0;
        error = 0;
        eintegral = 0;
        ederivative = 0;
        for(int d=0; d++; d<100){
          checkSpeed[d] = 0;
        }

        rotationSpindle();
        if(enb_stop == 1){ //normal conclusion of the workcycle
          Stop();
        }else if(startState == LOW || secCheckState == LOW){ //emergency stop of the workcycle
          Emergency();
          return;
        }
      }else if(emergencyState == HIGH){
        Emergency();
        return;
      }
    }
  }
}

int receiveEnableStart(){
  receiveOK = 0;
  enb = 0;
  do{
    if(Serial3.available()>0) { // Check if data is available to read
      String receivedString = Serial3.readStringUntil('\n'); // Read the incoming data until newline character
      enb = receivedString.toInt(); // Convert the received string to an integer for the target speed of the motor
      if(enb == 1){ //check if the speed can be accepted or not, decide a minimum and a maximum value
        receiveOK = 1;
        Serial.println(enb);  
      }
    }
    startState = digitalRead(startPin);
  }while(startState == HIGH && receiveOK == 0);

  delay(500);
  return enb;
}

int receiveSpeed(){
  receiveOK = 0;
  spd = 0;
  do{
    if(Serial3.available()>0) { // Check if data is available to read
      String receivedString = Serial3.readStringUntil('\n'); // Read the incoming data until newline character
      spd = receivedString.toInt(); // Convert the received string to an integer for the target speed of the motor
      if((spd > 0) && (spd <= 280)){ //check if the speed can be accepted or not, decide a minimum and a maximum value
        enb_spd = 1;
        Serial3.println(enb_spd);
        receiveOK = 1;
      }
    }
    startState = digitalRead(startPin);
  }while(startState == HIGH && receiveOK == 0);

  Serial.print("Received speed: ");
  Serial.println(spd);

  lcdSpeed.clear();
  lcdSpeed.print("Speed acquired");
  lcdSpeed.setCursor(0,1);
  lcdSpeed.print("correctly");
  delay(1000);
  return spd;
}

int receiveEnableRotation(){
  receiveOK = 0;
  enb_rot = 0;
  do{
    if(Serial3.available()>0) { // Check if data is available to read
      String receivedString = Serial3.readStringUntil('\n'); // Read the incoming data until newline character
      enb_rot = receivedString.toInt(); // Convert the received string to an integer for the target speed of the motor
      if(enb_rot == 1){ //check if the speed can be accepted or not, decide a minimum and a maximum value
        receiveOK = 1;
      }
    }
    startState = digitalRead(startPin);
    secCheckState = digitalRead(secCheckPin);
  }while(startState == HIGH && secCheckState == HIGH && receiveOK == 0);

  Serial.println(enb_rot);
  delay(1000);
  return enb_rot;
}

void rotationSpindle(){
  enb_stop = 0;

  digitalWrite(startLed,HIGH); //switch on the start led (ON and window closed)
  digitalWrite(pauseLed,HIGH); //keep the window led on
  digitalWrite(stopLed,HIGH);

  do{
    startState = digitalRead(startPin); //acquisition of the signal from the start
    secCheckState = digitalRead(secCheckPin); //acquisition of the signal from the securuty window over the spindle

    if(startState == HIGH && secCheckState == HIGH){
      //read the position in an atomic block to avoid potential misreads
      pos = 0;
      ATOMIC_BLOCK(ATOMIC_RESTORESTATE){
        pos = pos_i;
      }

      //method 1: count the number of pulses within a time unit
      //compute the time between each iteration of the loop function
      currT = micros(); //measure the time elapsed
      deltaT = ((float) (currT - prevT))/1.0e6; //time interval
      vel = (pos - posPrev)/deltaT; //encoder counts per second
      posPrev = pos;
      prevT = currT;

      //convert count/s to RPM and the frequency compensation for the PWM frequency variation
      velocity = vel/600.0*60.0;
      
      //Low pass first order filter (10 Hz cut-off)
      velFilt = 0.9391*velFilt1 + 0.03046*velocity + 0.03046*velPrev1;
      velPrev1 = velocity;
      velFilt1 = velFilt;

      velFiltInt = (int)velFilt;

      j++;
      if(j == 99){
        lcdSpeed.clear();
        lcdSpeed.print("Speed: ");
        lcdSpeed.setCursor(0,1);
        lcdSpeed.print(String(velFiltInt) + " rpm");
        j = 0;
      }

      //compute the control signal u
      kp = 2.0;
      ki = 0.13;
      kd = 12;

      u = PIDcontroller(targetSpeed, kp, ki, kd);
      setMotor(u,pwmPin);

      Serial.print(deltaT);
      Serial.print("\t");
      Serial.print(targetSpeed);
      Serial.print("\t");
      Serial.print(MinimumTargetSpeed);
      Serial.print("\t");
      Serial.print(MaximumTargetSpeed);
      Serial.print("\t");
      Serial.print(velFiltInt);
      Serial.print("\t");
      Serial.print(u);
      Serial.print("\t");
      Serial.print(Pout);
      Serial.print("\t");
      Serial.print(Iout);
      Serial.print("\t");
      Serial.println(Dout);

      //check of the speed to send to the Arduino the real spindle speed
      if(velFiltInt >= MinimumTargetSpeed && velFiltInt <= MaximumTargetSpeed && snd == 0){
        checkSpeed[i] = velFiltInt;
        i++;
        if(i == 99){
          realSpeed = velFiltInt;
          snd = 1;
          Serial.println("");
          Serial.print("\t \t");
          Serial.println(realSpeed);
          Serial.println("");
          sendSpeed();
        }
      }else if((velFiltInt <= MinimumTargetSpeed || velFiltInt >= MaximumTargetSpeed) && snd == 0){
        for(int d=0; d++; d<100){
          checkSpeed[d] = 0;
        }          
        i = 0;
      }
      
      delay(1); //according to the sampling time
    }
    
    if(Serial3.available()>0){ // Check if data is available to read
      String receivedStop = Serial3.readStringUntil('\n'); // Read the incoming data until newline character
      enb_stop = receivedStop.toInt(); // Convert the received string to an integer
    }
  }while(startState == HIGH && secCheckState == HIGH && enb_stop == 0);
}

void Stop(){ //stop the rotation of the spindle when the process of turning is finished
  targetSpeed = 0;
  u = 0;
  setMotor(u,pwmPin); //stop the motor

  digitalWrite(emergencyLed,LOW);
  digitalWrite(stopLed,HIGH); 
  digitalWrite(pauseLed,LOW); //turn on the blue-stop light
  digitalWrite(startLed,LOW); //turn off the green-start light

  lcdSpeed.clear();
  lcdSpeed.home();
  lcdSpeed.print("The process");
  lcdSpeed.setCursor(0,1);
  lcdSpeed.print("is completed");

  realSpeed = 0;
  velFiltInt = 0;
  for(int d=0; d++; d<100){
    checkSpeed[d] = 0;
  } 

  do{
    secCheckState = digitalRead(secCheckPin);
    delay(100);
  }while(startState == HIGH && secCheckState == HIGH);
}

void Emergency(){ //stop the motor, switch on the emergency/fault led
  targetSpeed = 0;
  u = 0;
  setMotor(u,pwmPin); //stop the motor

  digitalWrite(emergencyLed,HIGH);
  digitalWrite(startLed,LOW); //switch on the start led (ON and window closed)
  digitalWrite(pauseLed,LOW); //keep the window led on
  digitalWrite(stopLed,LOW);
  
  lcdSpeed.clear();
  lcdSpeed.home();
  lcdSpeed.print("EMERGENCY");
  lcdSpeed.setCursor(0,1);
  lcdSpeed.print("STOP");

  Serial.println("EMERGENCY STOP");

  realSpeed = 0;
  velFiltInt = 0;
  for(int d=0; d++; d<100){
    checkSpeed[d] = 0;
  }

  do{
    emergencyState = digitalRead(emergencyStopPin);
    delay(100);
  }while(emergencyState == HIGH);
}

void sendSpeed(){
  snd = 1; //the transmission of the speed is performed one time
  Serial3.println(realSpeed);
}

//function to determine the direction of rotation of the motor shaft
void readEncoder(){
  b = digitalRead(ENCB); //acquire the second signal when the first triggers
  increment = 0;
  if(b>0){
    increment = 1; //forward direction (clockwise)
  }else{
    increment = -1; //backward direction (counter-clockwise)
  }
  pos_i = pos_i + increment;
}

float PIDcontroller(int target, float KP, float KI, float KD){
  error = (float) (targetSpeed - velFiltInt);
  
  eintegral = eintegral + error*deltaT;
  ederivative = (error - lastError)/deltaT;
  Pout = KP*error;
  Iout = KI*eintegral;

  if(Iout < LowerLimit_Iout){
    Iout = LowerLimit_Iout;
  }else if(Iout > UpperLimit_Iout){
    Iout = UpperLimit_Iout;
  }

  Dout = KD*ederivative;

  if(deltaT <= sampleTime){
    u = Pout + Iout; + Dout;
  }else if(deltaT > sampleTime){
    u = Pout;
    eintegral = 0;
  }

  lastError = error;
  return u;
}

//function to change the frequency of the PWM signal
void setPwmFrequency(int pin, int divisor){
  byte mode;
    switch(divisor){ //based on the desired divisor, the variable mode assumes an hexadecimal value
      case 1: mode = 0x01; break;
      case 8: mode = 0x02; break;
      case 32: mode = 0x03; break;
      case 64: mode = 0x04; break;
      case 128: mode = 0x05; break;
      case 256: mode = 0x06; break;
      case 1024: mode = 0x07; break;
      default: return;
    }

    switch(pin){
      case 2:  TCCR3B = TCCR3B  & 0b11111000 | mode; break;
      case 3:  TCCR3B = TCCR3B  & 0b11111000 | mode; break;
      case 4:  TCCR0B = TCCR0B  & 0b11111000 | mode; break;
      case 5:  TCCR3B = TCCR3B  & 0b11111000 | mode; break;
      case 6:  TCCR4B = TCCR4B  & 0b11111000 | mode; break;
      case 7:  TCCR4B = TCCR4B  & 0b11111000 | mode; break;
      case 8:  TCCR4B = TCCR4B  & 0b11111000 | mode; break;
      case 9:  TCCR2B = TCCR0B  & 0b11111000 | mode; break;
      case 10: TCCR2B = TCCR2B  & 0b11111000 | mode; break;
      case 11: TCCR1B = TCCR1B  & 0b11111000 | mode; break;  
      case 12: TCCR1B = TCCR1B  & 0b11111000 | mode; break;  
      case 13: TCCR0B = TCCR0B  & 0b11111000 | mode; break;
      default: return;
    }
}

//function to drive the motor
void setMotor(float input, int PinMotor){
  int pwmVal = (int) fabs(input);
  if(pwmVal > 255){
    pwmVal = 255;
  }
  analogWrite(PinMotor,pwmVal); //set the speed of the motor
}