//ON-OFF and stop signals
#define startPin 34 //pin for the start signal
bool startState = 0; //variable for the state of the start signal

#define secCheckPin 36 //pin for the security window over the spindle
bool secCheckState = 0; //variable for the state of the security window over the spindle

bool stopState = 0; //variable for the state of the stop signal

//LCD display with I2C serial communication
#include <Wire.h> //this arduino is the Master
#include <LiquidCrystal_I2C.h>

#define I2C_ADDR 0x27
#define BACKLIGHT_PIN 3 //default pins for the serial communication
#define En_pin 2
#define Rw_pin 1
#define Rs_pin 0
#define D4_pin 4
#define D5_pin 5
#define D6_pin 6
#define D7_pin 7

LiquidCrystal_I2C lcd(I2C_ADDR,En_pin,Rw_pin,Rs_pin,D4_pin,D5_pin,D6_pin,D7_pin);

//Membrane keypad definition
#include <Keypad.h>
const byte rows = 4; //constants for the rows and for the columns
const byte cols = 4;

//Connections to Arduino board
byte rowPins[rows] = {69, 68, 67, 66};
byte colPins[cols] = {65, 64, 63, 62};

//Matrix to represent the keys on keypad
char hexaKeys[rows][cols] = {
  {'1', '2', '3', 'A'},
  {'4', '5', '6', 'B'},
  {'7', '8', '9', 'C'},
  {'*', '0', '#', 'D'}
};
//Create the keypad object
Keypad customKeypad = Keypad(makeKeymap(hexaKeys), rowPins, colPins, rows, cols);

char varStart; //variable to start the process
char varMode; //variable to choose the operating mode
int ok = 0;

//definition of the limits for the working parameters
float maximumDiameter = 30.00; //maximum acceptable working diameter of the piece [mm]
int maximumCuttingSpeed = 40; //maximum working speed [m/min]
int maxSpindleSpeed = 280; //maximum spindle rotational speed [rpm]
float maxLength = 200.00; //maximum length of the machine (admissible for the tool movement) [mm]
float maxFeedX = 0.5; //maximum value allowed for the feed along the X axis

//definition of the working parameters
float startDiameter = 0.00; //starting diameter
String diameter1 = "";
float finalDiameter = 0.00; //final diameter
String diameter2 = "";
bool checkDiameter = 0; //variable to check the correctness of the starting diameter

float feedXint = 0;
float feedX = 0;
float feedZ = 0;
int numberCycles = 0; //number of stages of work
float feedXrough = 0.00; //depth of one cycle
float feedZrough = 0.00; //feed rate along the z-axis for the roughing process

float feedXfinish = 0.010; //feed X for the finishing process
float feedZfinish = 0;

float feedX1 = 0.00; //depth of one cycle
float feedX2 = 0; //depth of the finish process

float finalDiameterBegin = 0;
float finalDiameterEnd = 0;
float avgFinalDiameter = 0;
float feedXbegin = 0;
int NmaxBegin = 0;
float feedXend = 0;
int NmaxEnd = 0;
float feedXcyclindrical = 0;
float interDiam = 0;
bool conicity = 0;

int cuttingSpeed = 0; //cutting speed selected
bool checkSpeed = 0;
String speed = "";
const float piGreco = 3.14;
bool checkSpindleSpeed = 0;
int spindleSpeed = 0;
float toleranceSpeed = 0.02;
int targetMin = 0;
int targetMax = 0;

float feedZ1 = 0.00; //feed rate along the z-axis for the roughing process
float feedZ2 = 0.00; //feed rate along the z-axis for the finish process
bool checkFeedZ = 0.00;
String movementZ = "";
float maximumFeedRateZ = 0.3;
int rotationalSpeedScrew = 0;

int okSend = 0;
String receivedSpeed = ""; //String to hold the incoming speed from the other Arduino board
int realSpindleSpeed = 0; //variable to save the real speed of the spindle from the slaver
bool okSpeed = 0;
int controlSpeed = 0;

char varEnable; //variable to receive the enabling for the start of the rotation of the spindle
bool okEnable = 0;

//ultrasonic sensor useful to measure the presence and distance of the object to work with respect to the home position
#define trigPin 22 //generates the soundwave to send to the object
#define echoPin 23 //listen to the soundwave coming back from the object

//variables to store the time interval between the soundwaves and the relative distance (knowing the speed of the sound in the air)
float duration = 0;
float distance = 0;

//compute the distance home position-object computing the average among 100 measurements
float measures[101];
float sum = 0;
float averageDistance = 0;

const float extra_len = 5; //extra-distance from the object during the positioning of the tool [mm]
float positioning = 0; //store the actual distance for the tool positioning before performing the work
float pitch = 1.00; //pitch of the screw of the Z-axis and X-axis [mm]

float tot_distance = 0;

//definition of the connections for the A4988 driver for the stepper motor Z-axis movement
#define dirPin 26 //pin for the motion direction
#define stepPin 28 //pin for the regulation of the speed of the pulses train
#define enPin 30 //pin for the enable of the driver
#define MS1 27
#define MS2 29
#define MS3 31

int NpulsesRevolutionZ = 0; //number of pulses to send to the motor to perform one revolution Z-axis
const float stepAngle = 1.8; //step angle of the stepper motor [°]
int microstepZ = 800; //microsteps for the motor NEMA17
float T_screw = 0;
int halfT_screw = 0;

//connections of the ULN2003A driver for the stepper motor X-axis movement
#define in1 9 //phase 1 blue
#define in2 10 //phase 2 pink
#define in3 11 //phase 3 yellow
#define in4 12 //phase 4 orange
int delayMotor = 0;

int NpulsesRevolutionX = 512; //full mode for the motor X
const int gearRatio = 64; //gear ratio of the stepper motor in the X-axis
const int angleX = 45; //angle covered by the X-axis motor when a cycle of activation of the 4 phases

float positionXaxis = 0;
float xHome = 5;
long pulsesPosX = 0;

float workingLength = 0;
bool checkWorkingLength = 0;
String workDistance = "";
bool checkOverallDistance = 0;

float distanceCovered = 0.00;

float distanceZ = 0;
long pulsesZ = 0; //number of pulses to send to the stepper motor
long countZ = 0;

float distanceX = 0;
long pulsesX = 0; //number of pulses to send to the stepper motor
long countX = 0;

long pulses = 0; //number of pulses to send to the stepper motor
long count = 0;

int nC = 0;

float halfperiodSpeed = 0;
int periodSpeed = 0;

volatile long comebackVarZ = 0; //variable used to place back the tool after the process, along the Z-axis. It contains the total number of pulses sent to the motor to perform both the positioning and the work process
volatile long comebackVarX = 0; //variable used to place back the tool after the process, along the X-axis
long internVarZ = 0;
long internVarX = 0;

int checkConicity = 0;
float residualFeedXbegin = 0;
float residualFeedXend = 0;
int numberCyclesCyl = 0;
float deltaX = 0;
float deltaZ = 0;
float feedXconicalBegin = 0;
float feedXconicalEnd = 0;
int NcyclesConicalMax = 0;

long pulsesZconical = 0;
long pulsesXconical = 0;
float partialPulses_float= 0;
float residualPulses_float = 0;
int partialPulses= 0;
int residualPulses = 0;

int incr = 0;
int incrRes = 0;
int n = 0;
int pulsesExtra = 0;
int dx = 0;

float finalDiameter1 = 0;
float finalDiameter2 = 0;

int nCycles1 = 0;
int nCycles2 = 0;

float workingLengthTotal = 0;
float workingLength1 = 0;
float workingLength2 = 0;

//emergency stop
#define emergencyPin 19 //pin for the signal generating the interrupt
bool emergencyState = false; //it becomes true when the code enters the emergency function/state

//settings for the tft display ST7796S
#include <LCDWIKI_GUI.h> // Core graphics library
#include <LCDWIKI_SPI.h> // Hardware-specific library

//paramters define
#define MODEL ST7796S
#define CS   A5    
#define CD   A3
#define RST  A4
#define LED  A0

// This will instantiate (i.e. create) a new LCDWIKI_SPI object
// that you will be able to interact with
LCDWIKI_SPI mylcd(MODEL, CS, CD, RST, LED); 

// Assign human-readable names to some common 16-bit color values:
#define	BLACK   0x0000
#define	BLUE    0x001F
#define	RED     0xF800
#define	GREEN   0x07E0
#define CYAN    0x07FF
#define MAGENTA 0xF81F
#define YELLOW  0xFFE0
#define WHITE   0xFFFF

bool speedSignalState = 0;
int enb = 0;
int enb_spd = 0;
int enb_rot = 0;
bool receiveOK = 0;
int enb_stop = 0;

float ver = 0;

char varBack;
bool varBackstate = 0;

void setup(){
  Serial.begin(9600);
  Serial3.begin(9600);

  lcd.begin(16,2);
  lcd.clear();
  lcd.setBacklightPin(BACKLIGHT_PIN,POSITIVE);
  lcd.setBacklight(HIGH);

  pinMode(startPin,INPUT_PULLUP); //enabling of the pin for the reading of the start
  pinMode(secCheckPin,INPUT_PULLUP);

  pinMode(emergencyPin, OUTPUT);
  emergencyState = false;

  pinMode(trigPin, OUTPUT);
  pinMode(echoPin, INPUT);

  pinMode(stepPin,OUTPUT); 
  pinMode(dirPin,OUTPUT);
  pinMode(enPin,OUTPUT);
  digitalWrite(enPin,LOW);
  pinMode(MS1,OUTPUT);
  pinMode(MS2,OUTPUT);
  pinMode(MS3,OUTPUT);
  digitalWrite(MS1,LOW);
  digitalWrite(MS2,HIGH);
  digitalWrite(MS3,LOW);
  NpulsesRevolutionZ = 800; //number based on the combination of the 3 pins above
  T_screw = 60*1e6;
  T_screw = T_screw/2;
  T_screw = T_screw/200;
  T_screw = T_screw/NpulsesRevolutionZ;
  halfT_screw = (int)T_screw;

  pinMode(in1,OUTPUT);
  pinMode(in2,OUTPUT);
  pinMode(in3,OUTPUT);
  pinMode(in4,OUTPUT);
  digitalWrite(in1,LOW);
  digitalWrite(in2,LOW);
  digitalWrite(in3,LOW);
  digitalWrite(in4,LOW);  

  delay(3000); //to avoid any misbehaviour when the code is uploaded on the Arduino board
  
  mylcd.Init_LCD();
  mylcd.Set_Rotation(1); //0 for portraint, 1 for landscape
  mylcd.Fill_Screen(BLUE); //set the colour of the background
  mylcd.Set_Text_Mode(1); //set the orientation of the text on the display
  mylcd.Set_Text_colour(WHITE); //set the clolour of the text
  mylcd.Set_Text_Back_colour(BLUE); //set the clolour of the text background
  mylcd.Set_Text_Size(2); //size of the text
}

void loop(){
  startState = digitalRead(startPin); //acquisition of the signal from the start connection
  secCheckState = digitalRead(secCheckPin);
  
  if(startState == HIGH && secCheckState == LOW){ //ON state
    mylcd.Print_String("Press the command # on the keypad", 0, 0); //print this text in the specified position
    mylcd.Print_String("to start the process", 0, 18);
    delay(50);

    varStart = customKeypad.getKey(); //acquisition of the command from the keypad
    enb = 0;
    enb_spd = 0;
    enb_rot = 0;
    receiveOK = 0;
    enb_stop = 0;

    if(varStart && varStart == '#'){
      mylcd.Fill_Screen(BLUE);
      mylcd.Print_String("Choose a working process ", 0, 0);
      mylcd.Print_String("from the following list:", 0, 18);
      mylcd.Print_String("> A: complete cylindrical turning with", 0, 36);
      mylcd.Print_String("roughing + finishing stages", 0, 54);
      mylcd.Print_String("> B: cylindrical turning for a", 0, 72);
      mylcd.Print_String("multidiameter component with", 0, 90);
      mylcd.Print_String("roughing + finishing stages", 0, 108);
      mylcd.Print_String("> C: conical turning with", 0, 126);
      mylcd.Print_String("roughing + finishing stages", 0, 144);
      delay(200);

      comebackVarZ = 0;
      comebackVarX = 0;

      do{
        ok = 0;
        delay(50);
        startState = digitalRead(startPin);
        varMode = customKeypad.getKey();

        if(varMode && startState == HIGH){ //this is necessary because the varMode variable contains something, otherwise in the switch function the default branch is always selected
          switch(varMode){
            case 'A': lcd.clear();
              ok = completeProcess();
              lcd.clear();              lcd.print("Ready for a new");              lcd.setCursor(0,1);              lcd.print("process");
              delay(1000);
              mylcd.Fill_Screen(BLUE);
              lcd.clear();
            break;
            case 'B': lcd.clear();
              ok = multidiameter();
              lcd.clear();              lcd.print("Ready for a new");              lcd.setCursor(0,1);              lcd.print("process");
              delay(1000);
              mylcd.Fill_Screen(BLUE);
              lcd.clear();
            break;
            case 'C': lcd.clear();
              ok = conical();
              lcd.clear();              lcd.print("Ready for a new");              lcd.setCursor(0,1);              lcd.print("process");
              delay(1000);
              mylcd.Fill_Screen(BLUE);
              lcd.clear();
            break;
            default: 
              lcd.clear();              lcd.print("Insert not valid");
              mylcd.Print_String("Insert not valid", 0, 216);
              delay(1000);
              mylcd.Fill_Screen(BLUE);
              lcd.clear();
            return;
          }
        }else if(startState == LOW){
          Emergency();
          ok = 2;
        }
      }while(ok == 0);
    }else if(varStart && varStart != '#'){
      lcd.clear();      lcd.print("Insert not valid");
      mylcd.Print_String("Insert not valid", 0, 54);
      delay(1000);
      mylcd.Fill_Screen(BLUE);
    }
  }else if(startState == LOW || secCheckState == HIGH){ //OFF state
    lcd.clear();
    mylcd.Fill_Screen(BLUE);
  }
}

void Emergency(){ //EMERGENCY STOP AND THE TOOL COMES BACK TO THE HOME POSITION
  emergencyState = !emergencyState; //the state becomes true 
  digitalWrite(emergencyPin, emergencyState); //send the info to the other Arduino

  mylcd.Fill_Screen(BLUE);
  mylcd.Print_String("Emergency stop", 0, 0);
  lcd.clear();  lcd.print("STOP");
  delay(1000);

  if(comebackVarX != 0 || comebackVarZ != 0){ //carriage already moved from the rest position
    mylcd.Print_String("Press the 'D' command", 0, 36);
    mylcd.Print_String("to move back the tool", 0, 54);

    varBackstate = 0;
    do{
      varBack = customKeypad.getKey();
      delay(50);

      if(varBack && varBack == 'D'){
        varBackstate = 1;
      }
    }while(varBackstate == 0);

    mylcd.Print_String("Moving back the tool", 0, 90);
    lcd.setCursor(0,1);  lcd.print("Moving back...");
    //acquire the position of the motors and move back the tool
    for(int i=0; i<comebackVarX; i++){
      Backward();
    }

    digitalWrite(dirPin,LOW);
    for(int j=0; j<comebackVarZ; j++){
      digitalWrite(stepPin,HIGH);
      delayMicroseconds(halfT_screw);
      digitalWrite(stepPin,LOW); 
      delayMicroseconds(halfT_screw);
    }

    comebackVarZ = 0;
    comebackVarX = 0;
    
    mylcd.Print_String("Open the security window", 0, 126);
    delay(1000);
    do{
      secCheckState = digitalRead(secCheckPin);
      delay(100);
    }while(secCheckState == HIGH);
  }else{ //emergency function during the setup procedure
    comebackVarZ = 0;
    comebackVarX = 0;
    
    mylcd.Print_String("Open the security window", 0, 36);
    delay(1000);
    do{
      secCheckState = digitalRead(secCheckPin);
      delay(100);
    }while(secCheckState == HIGH);
  }  

  emergencyState = !emergencyState; //the state becomes false
  digitalWrite(emergencyPin,emergencyState);
  mylcd.Fill_Screen(BLUE);
}

void backHome(long backZ, long backX){
  mylcd.Fill_Screen(BLUE);
  mylcd.Print_String("Work cycle completed", 0, 0);
  mylcd.Print_String("Moving back the tool:", 0, 18);
  lcd.clear();  lcd.print("Moving back...");
  delay(1000);
  Serial.print(backX);
  Serial.print("\t");
  Serial.println(backZ);
  delay(1000);

  //before back movement along the X-axis
  mylcd.Print_String("along the X-axis", 0, 36);
  lcd.clear();
  delayMotor = 2;
  for(int i=0; i<backX; i++){
    Backward();
    comebackVarX--;
    countX--;
    if(countX == 0){
      distanceX = (float) comebackVarX/NpulsesRevolutionX;
      countX = NpulsesRevolutionX;
      lcd.setCursor(0,1);      
      lcd.print("X pos: " + String(distanceX)  + " mm");
    }
  }

  //after back movement along the Z-axis
  mylcd.Print_String("along the Z-axis", 0, 54);
  digitalWrite(dirPin,LOW);
  for(int i=0; i<backZ; i++) {
    digitalWrite(stepPin,HIGH);
    delayMicroseconds(halfT_screw);
    digitalWrite(stepPin,LOW); 
    delayMicroseconds(halfT_screw);
    comebackVarZ--;
    countZ--;
    if(countZ == 0){
      distanceZ = (float) comebackVarZ/NpulsesRevolutionZ;
      countZ = NpulsesRevolutionZ;
      lcd.home();
      lcd.print("Z pos: " + String(distanceZ) + " mm");
    }
  }
  delay(1000);

  lcd.clear();  lcd.print("Cycle completed");  lcd.setCursor(0,1);  lcd.print("correctly");

  mylcd.Print_String("Process completed correctly", 0, 90);
  delay(1000);
}

void writeMotorX(bool a, bool b, bool c, bool d){
  digitalWrite(in1,a);
  digitalWrite(in2,b);
  digitalWrite(in3,c);
  digitalWrite(in4,d);
}

void Backward(){
  //Step 1
  writeMotorX(HIGH,LOW,LOW,LOW);
  delay(delayMotor);
  //Step 2
  writeMotorX(LOW,HIGH,LOW,LOW);
  delay(delayMotor);
  //Step 3
  writeMotorX(LOW,LOW,HIGH,LOW);
  delay(delayMotor);
  //Step 4
  writeMotorX(LOW,LOW,LOW,HIGH);
  delay(delayMotor);
}

void Forward(){
  //Step 4
  writeMotorX(LOW,LOW,LOW,HIGH);
  delay(delayMotor);
  //Step 3
  writeMotorX(LOW,LOW,HIGH,LOW);
  delay(delayMotor);
  //Step 2
  writeMotorX(LOW,HIGH,LOW,LOW);
  delay(delayMotor);
  //Step 1
  writeMotorX(HIGH,LOW,LOW,LOW);
  delay(delayMotor);
}